// $License: NOLICENSE
//--------------------------------------------------------------------------------
/**
  @file
  This file contains functions to perform actions needed to initialize installation tool.
  They initialize necessary global variables as well as perform upgrade of internal datapoints if needed.
*/

// Purposes:
// * Initialization of needed global variables each time the manager starts
// * Checking if initialization or upgrade is needed and performing respective actions

//--------------------------------------------------------------------------------
// used libraries (#uses)
#uses "fwInstallation/fwInstallation.ctl" // This whole library should not be loaded, needed functionality should be moved to some kind of utils library. fwInstallation.ctl (or other lib) should load this one.

//--------------------------------------------------------------------------------
// variables and constants

/** Version of this library.
 * Used to determine the coherency of all libraries of the installation tool
 * Please do not edit it manually
 * @ingroup Constants
*/
const string csFwInstallationInitLibVersion = "8.4.1";
const string csFwInstallationInitLibTag = "";

/** Name of the init file loaded at start up of the tool.
 * @ingroup Constants
*/
const string gFwInstallationInitScript = "fwInstallationInitScript.ctl";

//--------------------------------------------------------------------------------
//@public members
//--------------------------------------------------------------------------------

/** This function needs to be called before the first use of the installation library and after each installation.
  @param runPostInstall (in) this variable specifies whether pending post-install scripts, if any, must be run during initialization
  @return 0 if OK, -1 if errors
*/
int fwInstallationInit_execute(bool runPostInstall = true, string installationPath = ""){
  dyn_string dataPointTypes;
  int dpCreateResult;
  dyn_string dynPostInstallFiles_all; // all the postInstall init files to be executed
  dyn_string dynScripts;

  int iReturn;
  dyn_float dreturnf;
  dyn_string dreturns;
  string testString, startMode;
  string dp = fwInstallation_getInstallationDp();
  string dpa = fwInstallation_getAgentDp();

  //Add libs to config file if they do not exist yet
  fwInstallationInit_addLibToConfig();

  if ( !globalExists("gInstallationAgentUiNum") )
    addGlobal("gInstallationAgentUiNum", INT_VAR);

  if ( !globalExists("gFwInstallationLog") )
    addGlobal("gFwInstallationLog", DYN_DYN_STRING_VAR);

  //Initialize the PMON variables:
  if ( !globalExists("gFwInstallationPmonUser") )
    addGlobal("gFwInstallationPmonUser", STRING_VAR);

  if ( !globalExists("gFwInstallationPmonPwd") )
    addGlobal("gFwInstallationPmonPwd", STRING_VAR);

  if ( !globalExists("gFwInstallationCurrentComponent") )
    addGlobal("gFwInstallationCurrentComponent", STRING_VAR);

  gFwInstallationPmonUser = "N/A";
  gFwInstallationPmonPwd = "N/A";

  // check whether the _FwInstallationComponents dpt exists
  dataPointTypes = dpTypes();
  if (dynContains(dataPointTypes, "_FwInstallationComponents") <= 0)
  {
    fwInstallation_throw("Starting the Installation Tool for the first time", "INFO", 10);
      // create the installation tool internal data points
      fwInstallationInit_createDataPointTypes(TRUE, dpCreateResult);
      // check the result
      if (dpCreateResult)
        fwInstallation_throw("Failed to create the internal data points. Please re-install the FW Component Installation Tool");

    if(fwInstallationInit_createPendingActionsDataPointType())
    {
        fwInstallation_throw("Failed to create the internal data point type for pending actions. Please re-install the FW Component Installation Tool");
    }

    if(fwInstallationInit_createPendingActionsDp())
    {
        fwInstallation_throw("Failed to create the internal data point for pending actions. Please re-install the FW Component Installation Tool");
    }

   fwInstallation_throw("Internal data points for the installation tool created", "INFO", 10);
  }
  else if((fwInstallationRedu_isSplitMode() || !fwInstallationRedu_isPassive()) && fwInstallationInit_updateDPT())//Check that the DPTs are up-to-date:
  {
    fwInstallation_throw("Could not update internal DPTs of the installation tool");
    return -1;
  }
  else if(fwInstallationRedu_isRedundant())
  {//When redundant system is migrated from version < 8.1.0 ensure that new _FwInstallationComponents dpe .sourceFilesHases contains files hashes
    int emptyFilesHashes = 0;
    dyn_string installedComponentDps = fwInstallation_getInstalledComponentDps();
    int installedComponentDpsLen = dynlen(installedComponentDps);
    for(int i=1;i<=installedComponentDpsLen;i++)
    {
      string dpe = installedComponentDps[i] + ".sourceFilesHashes";
      if(!dpExists(dpe))
      {
        fwInstallation_throw("Internal DPTs of the installation tool are not up-to-date. Need to run on active peer first", "WARNING", 10);
        break;
      }
      dyn_string sourceFilesHashes;
      dpGet(dpe, sourceFilesHashes);
      if(dynlen(sourceFilesHashes) == 0)
      {
        emptyFilesHashes++;
      }
    }
    if(installedComponentDpsLen > 0 && emptyFilesHashes == installedComponentDpsLen)
    {//calculate hashes for currently installed components
      dyn_dyn_string installedComponents;
      if(fwInstallation_getInstalledComponents(installedComponents) != 0)
        fwInstallation_throw("Failed to get installed components properties. Cannot calculate hashes of source files. This needs to be done manually later", "ERROR");

      int installedComponentsCount = dynlen(installedComponents);
      for(int i=1;i<=installedComponentsCount;i++)
        if(fwInstallation_calculateComponentSourceFilesHashes(installedComponents[i][1]) != 0)
          fwInstallation_throw("Error when calculating source files hashes for component: " + installedComponents[i][1] +
                               ". Hash calculation needs for this component to be done manually later", "ERROR");
    }
  }

  if(!dpExists(dp))
  {
    dpCreate(dp, "_FwInstallationInformation");
    int count = 0;
    while(!dpExists(dp))
    {
      delay(0, 100);
      count++;
      if(count == 300)
        fwInstallation_throw("Failed to create internal fwInstallation dp: " + dp + " for 30 seconds. Cannot continue. Please kill the UI and check para privledges or contact support directly" , "ERROR");
    }
  }

  //make default option not to add managers on redu system:
  fwInstallationInit_setAddManagersOnReduPartner(false);

  //if there are components installed, make sure that the dp-element 'name' is properly filled:
  fwInstallationInit_updateComponentDps();

  // Installation path was given as an argument - set a DP, it will be later checked
  if(installationPath != "")
  {
    dpSet(dp + ".installationDirectoryPath", installationPath);
  }

  //Create trash if it does not exist:
  string sourceDir;
  dpGet(dp + ".installationDirectoryPath", sourceDir);

  if(sourceDir != "" && access(sourceDir, F_OK) >= 0)
  {
    fwInstallation_addProjPath(sourceDir, 999);
  }


  //Install installation agent:
  if(!fwInstallationInit_isAgentInstalled())
  {
    fwInstallation_throw("Installation Tool DB-agent not up-to-date. Forcing update now...", "INFO", 10);
    if(fwInstallationInit_installAgent())
    {
      fwInstallation_throw("fwInstallationInit_execute() -> Could not install FW installation Agent");
      return -1;
    }
    fwInstallation_throw("FW Component Installation DB-agent successfully installed", "INFO", 10);
  }

  fwInstallationInit_updateToolVersionInDp();

  //Load init file as the user may have defined the schema owner there
  fwInstallationInit_loadInitFile();
  //Load init scripts
  if(fwInstallationInit_runInitScript())
  {
    fwInstallation_setToolStatus(false);
    fwInstallation_throw("There were errors executing the init script FW Component Installation Tool");
    return -1;
  }
  //Check if the DB is to be used and if so, upgrade system table to populate the event computer id column if required:
  int projectId = -1;
  if(fwInstallationDB_getUseDB() && fwInstallationDB_connect() == 0)
  {
    //check version of the schema
    string version = "";
    int getSchemaRetVal = fwInstallationDB_getSchemaVersion(version);
    bool compareSchemaRetVal = (getSchemaRetVal == 0)?fwInstallationDB_compareSchemaVersion():false;
    if(!compareSchemaRetVal)//disconnect DB as DB schema is wrong or has wrong version
    {
      fwInstallationDB_closeDBConnection();
      fwInstallationDB_setUseDB(false);

      if(getSchemaRetVal != 0)//schema not found
        fwInstallation_throw("fwInstallationInit_execute(): Failed to retrieve version of the DB schema. Please check if Configuration DB System Information schema is present in the DB");
      else//wrong schema version
        fwInstallation_throw("fwInstallationInit_execute(): Current DB schema version: " + version + " is lower than required: " + FW_INSTALLATION_DB_REQUIRED_SCHEMA_VERSION + ". Please upgrade the DB schema");
    }
    else
      fwInstallationDB_isProjectRegistered(projectId);
  }

  //add control manager for post installation scripts:
  string user, pwd, host = fwInstallation_getPmonHostname();
  int port = pmonPort();

  fwInstallation_getPmonInfo(user, pwd);

  //Add Agent to the PVSS console:
  fwInstallationManager_add(fwInstallation_getWCCOAExecutable("ctrl"), "always", 30, 3, 3,
                            FW_INSTALLATION_DB_AGENT_MANAGER_CMD, host, port, user, pwd);

  fwInstallationManager_add(fwInstallation_getWCCOAExecutable("ctrl"), "once", 30, 1, 1,
                            FW_INSTALLATION_SCRIPTS_MANAGER_CMD, host, port, user, pwd);
  //fwInstallation_throw("Ctrl Manager to project console for component post-install scripts added", "INFO", 10);

  if(runPostInstall)
  {
    fwInstallation_executePostInstallScripts();
  }

// add control manager for redundant systems:
//   if(fwInstallationRedu_isRedundant())
//   {
//     fwInstallationManager_add(fwInstallation_getWCCOAExecutable("ctrl"), "always", 30, 1, 1, "-f fwInstallationRedu.lst", host, port, user, pwd);
//   }


/*
  //Deploy the crash action script for the installation tool:
  if(VERSION != "3.6")
  {
    if(fwInstallation_deployCrashActionScript())
    {
      fwInstallation_setToolStatus(false);
      fwInstallation_throw("fwInstallationInit_execute() -> Failed to deploy the crash action script of the FW Component Installation Tool");
      return -1;
    }
  }
*/

//  fwInstallation_throw("*** FW Component Installation Tool v." + csFwInstallationToolVersion+ " ready ***", "INFO", 10);

  //Fill pvss version and patch list here:
  fwInstallationInit_updateWCCOAInfo();
  fwInstallation_setToolStatus(true);
  return 0;
}

/** This function provides the migration path from the fwInstallation versions < 8.4.0 to the latest ones.
  * It updates the fwInstallation LoadCtrlLibs config entries, removes obsolete files and restarts manager.
  * When it is called in UI manager, user has to confirm the actions.
  * Note: Due to the changes in the library locations, this function has to be called before fwInstallation script/panel checks the consistency of the library versions.
  *       It is because the existing old versions of libraries are already loaded via LoadCtrlLibs entry in the config file,
          and the new libraries, which were not present in the previous version are loaded with #uses statement.
*/
fwInstallationInit_consolidationCleanup(){
  string configPath = getPath(CONFIG_REL_PATH, "config");

  mapping obsoleteConfigLibs;
  fwInstallationInit_consolidation_getObsoleteConfigLibs(configPath, obsoleteConfigLibs);

  dyn_string filesToRemove, dirsToRemove, dirsToRemoveRecursively;
  fwInstallationInit_consolidation_getObsoletePathsToRemove(filesToRemove, dirsToRemove, dirsToRemoveRecursively);

  if(mappinglen(obsoleteConfigLibs) == 0 && dynlen(filesToRemove) == 0 &&
     dynlen(dirsToRemoveRecursively) == 0){
    return; // we've got there only because version number is not updated yet, no need for cleanup
  }

  fwInstallation_throw("fwInstallation will perform the update of its entries in project config file and obsolete files' cleanup", "INFO", 28);

  bool isLoadNoCtrlLibsOptionUsed = !isFunctionDefined("RootPanelOn"); // detect whether we use -LoadNoCtrlLib option to start manager
                                                                       // in this case the RootPanelOn function from panel.ctl is not defined
  if(myManType() == UI_MAN){
    if(isLoadNoCtrlLibsOptionUsed){
      execScript("#uses \"panel.ctl\" main(){}"); // needed as panels.ctl is not yet loaded
    }
    fwInstallation_throw("Waiting for user decision...", "INFO", 28);
    dyn_string ds;
    dyn_float df;
    ChildPanelOnCentralModalReturn("fwInstallation/fwInstallationUpgrade.pnl", "InstallationToolUpgrade",
                                   makeDynString("$obsoleteCfgLibs:" + fwInstallationInit_mappingObsoleteCfgLibsToString(obsoleteConfigLibs),
                                                 "$filesToRemove:" + strjoin(filesToRemove, "|"),
                                                 "$dirsToRemove:" + strjoin(dirsToRemove, "|"),
                                                 "$dirsToRmvRec:" + strjoin(dirsToRemoveRecursively, "|")), df, ds);
    if(dynlen(ds) <= 0 || ds[1] != "yes"){
      fwInstallation_throw("User decided not to continue the Installation Tool upgrade procedure. Closing the panel.", "WARNING", 28);
      PanelOff(); // close fwInstallation panel, but don't stop the manager
      delay(10); // ensure that panel will be closed, before we continue
      return;
    }
    fwInstallation_throw("User agreed to continue the Installation Tool upgrade procedure.", "INFO", 28);
  }

  if(fwInstallationInit_consolidation_removeObsoleteConfigLibs(configPath, obsoleteConfigLibs) != 0){
    fwInstallation_throw("Failed to remove obsolete fwInstallation entries from the config file. Cannot continue. Current manager will be stopped.", "ERROR", 28);
    exit(1);
  }

  int addCfgEntriesStatus = fwInstallationInit_addLibToConfig();

  fwInstallationInit_consolidation_removeObsoletePaths(filesToRemove, dirsToRemove, dirsToRemoveRecursively);

  fwInstallation_throw("---------------------------", "INFO", 28);
  fwInstallation_throw("Cleanup procedure finished.", "INFO", 28);

  
  if(addCfgEntriesStatus != 0){
    fwInstallationInit_notifyUserAndStopManager(
        "Failed to add required entries to the config file. It is not recommended to continue until this problem is fixed. Current UI manager will be stopped now.",
        "Required project config entries are missing. It is not recommended to continue until this problem is fixed. Current manager will be stopped now.",
        "ERROR", 1);
  }

  if(mappinglen(obsoleteConfigLibs) > 0){
    fwInstallationInit_notifyUserAndStopManager(
        "Installation Tool upgrade to version 8.4.0 finished.\nProject config file was updated. Obsolete libraries are still loaded by current UI manager. It is neccessary to load the new libraries.\n" +
        "Current UI manager will be stopped. Please restart it manually, if start mode is different than 'always'.\nIf -LoadNoCtrlLib option was used, do not activate it again.",
        "Project config file was updated. Current manager will be stopped, please restart it in order to continue. If -LoadNoCtrlLib option was used, do not activate it again.");
  }
  if(isLoadNoCtrlLibsOptionUsed){
    fwInstallationInit_notifyUserAndStopManager(
        "Installation Tool upgrade to version 8.4.0 finished.\nDetected, that UI manager was started with -LoadNoCtrlLib option. It is not recommended to continue operation, " +
        "having this option activated after the upgrade is finished.\nCurrent UI manager will be stopped. Please restart it manually without using -LoadNoCtrlLib option.",
        "Detected -LoadNoCtrlLib option for manager. Current manager will be stopped, please restart it without using -LoadNoCtrlLib option in order to continue");
  }
  if(!fwInstallationInit_consolidation_areLibsVersionsConsistent()){
    fwInstallationInit_notifyUserAndStopManager(
        "Installation Tool upgrade to version 8.4.0 finished.\nObsolete libraries are still loaded by current UI manager. It is neccessary to load the new libraries.\n" +
        "Current UI manager will be stopped. Please restart it manually, if start mode is different than 'always'.",
        "Current manager will be stopped to unload obsolete libraries, please restart it in order to continue.");
  }
}

/** Prints message to the log and in UI manager displays pop-up informing about necessity to stop manager. Then terminates the manager with given exit code.
  */
fwInstallationInit_notifyUserAndStopManager(string uiMessage, string logMessage, string messageType = "INFO", int exitCode = 0){
  fwInstallation_throw(logMessage, messageType, 28);
  if(myManType() == UI_MAN){
    dyn_string ds;
    dyn_float df;
    ChildPanelOnCentralModalReturn("fwInstallation/fwInstallation_messageInfo.pnl", "Manager restart required",
                                   makeDynString("$text:" + uiMessage, "$okLabel:Continue", "$hideCancelBtn:yes"), df, ds);
  }
  exit(exitCode);
}

/** This function checks whether a migration procedure has to be perfomed to upgrade fwInstallation from versions < 8.4.0 to the latest ones.
  */
bool fwInstallationInit_isCleanupNeeded(){
  if(!fwInstallationInit_consolidation_areLibsVersionsConsistent()){
    return true; // library versions are compared for the case when old libraries are already loaded - in this case cleanup procedure is obligatory
  }
  // additional checks needed when -LoadNoCtrlLib option is used, then comparison of library versions is not sufficient

  string fwInstallationDp = fwInstallation_getInstallationDp();
  if(dpExists(fwInstallationDp)){ // check if fwInstallation dp exists (is fwInstallation already initialized in the project)
    // look for obsolete config entries that are added by fwInstallation
    mapping obsoleteConfigLibs;
    fwInstallationInit_consolidation_getObsoleteConfigLibs(getPath(CONFIG_REL_PATH, "config"), obsoleteConfigLibs);
    if(mappinglen(obsoleteConfigLibs) > 0){
      return true;
    }
    // check previous fwInstallation version
    string fwInstallationVersion;
    dpGet(fwInstallationDp + ".version", fwInstallationVersion);
    if(fwInstallationVersion == ""){
      return true; // fwInstallation may be migrated from version, that did not set the 'version' dpe.
    }
    if(_fwInstallation_CompareVersions(fwInstallationVersion, "8.4.0", false, false, true) != 1){
      return true; //  fwInstallation version is lower than 8.4.0, cleanup needed
    }
  }
  return false; // version >= 8.4.0 or first run of fwInstallation
}

//--------------------------------------------------------------------------------
//@private members
//--------------------------------------------------------------------------------

/** This function adds the main libraries of the installation tool to the config file of the project
*/
int fwInstallationInit_addLibToConfig()
{
  dyn_string libs;
  bool err;

  paCfgReadValueList(PROJ_PATH + CONFIG_REL_PATH + FW_INSTALLATION_CONFIG_FILE_NAME, "ui", "LoadCtrlLibs", libs);
  if(dynContains(libs, "fwInstallation/fwInstallation.ctl") == 0){
    if(paCfgInsertValue(PROJ_PATH + CONFIG_REL_PATH + FW_INSTALLATION_CONFIG_FILE_NAME, "ui", "LoadCtrlLibs", "fwInstallation/fwInstallation.ctl") == 0){
      fwInstallation_throw("Added 'LoadCtrlLibs = \"fwInstallation/fwInstallation.ctl\"' entry to the [ui] section of the project config file", "INFO");
    }else{
      fwInstallation_throw("Failed to add 'LoadCtrlLibs = \"fwInstallation/fwInstallation.ctl\"' entry to the [ui] section of the project config file", "ERROR");
      err = true;
    }
  }

  paCfgReadValueList(PROJ_PATH + CONFIG_REL_PATH + FW_INSTALLATION_CONFIG_FILE_NAME, "ctrl", "LoadCtrlLibs", libs);
  if(dynContains(libs, "fwInstallation/fwInstallation.ctl") == 0){
    if(paCfgInsertValue(PROJ_PATH + CONFIG_REL_PATH + FW_INSTALLATION_CONFIG_FILE_NAME, "ctrl", "LoadCtrlLibs", "fwInstallation/fwInstallation.ctl") == 0){
      fwInstallation_throw("Added 'LoadCtrlLibs = \"fwInstallation/fwInstallation.ctl\"' entry to the [ctrl] section of the project config file", "INFO");
    }else{
      fwInstallation_throw("Failed to add 'LoadCtrlLibs = \"fwInstallation/fwInstallation.ctl\"' entry to the [ctrl] section of the project config file", "ERROR");
      err = true;
    }
  }
  if(err){
    fwInstallation_throw("Required config entries are missing. Components may work incorrectly after installation. " +
                         "Please manually add reported entries to the project config file or contact support.", "WARNING");
  }
  return err?-1:0;
}

/** This function creates the Installation Component DPT and DP
@param create (in) flag to indicate if an existing dp-type has to be overwritten (obsolete, legacy)
@param result (out) result of the operation, 0 if OK, -1 if error
@author M.Sliwinski, adapted by S. Schmeling and F. Varela.
*/
fwInstallationInit_createDataPointTypes(bool create, int & result)
{
  int n;
  dyn_dyn_string dynDynElements;
  dyn_dyn_int dynDynTypes;

  result = 0;

  int pos = 1; //Note: [pos++] - first returns pos value, then increments it.
  dynDynElements[pos] = makeDynString ("_FwInstallationComponents", "", "");
  dynDynTypes[pos++] =  makeDynInt (DPEL_STRUCT);

  dynDynElements[pos] = makeDynString ("", "componentFiles", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);
  dynDynElements[pos] = makeDynString ("", "configFiles", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_STRUCT);

  dynDynElements[pos] = makeDynString ("", "", "configWindows");
  dynDynTypes[pos++]  = makeDynInt    (0, 0, DPEL_DYN_STRING);
  dynDynElements[pos] = makeDynString ("", "", "configLinux");
  dynDynTypes[pos++]  = makeDynInt    (0, 0, DPEL_DYN_STRING);
  dynDynElements[pos] = makeDynString ("", "", "configGeneral");
  dynDynTypes[pos++]  = makeDynInt    (0, 0, DPEL_DYN_STRING);


  dynDynElements[pos] = makeDynString ("", "initFiles", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);
  dynDynElements[pos] = makeDynString ("", "postInstallFiles", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);

  dynDynElements[pos] = makeDynString ("", "dplistFiles", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);
  dynDynElements[pos] = makeDynString ("", "sourceFilesHashes", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);

  dynDynElements[pos] = makeDynString ("", "componentVersion", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_FLOAT);
  dynDynElements[pos] = makeDynString ("", "date", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_STRING);
  dynDynElements[pos] = makeDynString ("", "descFile", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_STRING);
  dynDynElements[pos] = makeDynString ("", "installationDirectory", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_STRING);
  dynDynElements[pos] = makeDynString ("", "requiredComponents", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);
  dynDynElements[pos] = makeDynString ("", "requiredInstalled", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_BOOL);
  dynDynElements[pos] = makeDynString ("", "subComponents", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);
  dynDynElements[pos] = makeDynString ("", "isItSubComponent", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_BOOL);
  dynDynElements[pos] = makeDynString ("", "scriptFiles", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);
  dynDynElements[pos] = makeDynString ("", "componentVersionString", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_STRING);

  dynDynElements[pos] = makeDynString ("", "deleteFiles", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);
  dynDynElements[pos] = makeDynString ("", "postDeleteFiles", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);

  dynDynElements[pos] = makeDynString ("", "helpFile", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_STRING);
  dynDynElements[pos] = makeDynString ("", "qtHelpFiles", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);

  dynDynElements[pos] = makeDynString ("", "sourceDir", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_STRING);
  dynDynElements[pos] = makeDynString ("", "installationNotOK", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_BOOL);
  dynDynElements[pos] = makeDynString ("", "comments", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_DYN_STRING);
  dynDynElements[pos] = makeDynString ("", "name", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_STRING);
  dynDynElements[pos] = makeDynString ("", "description", "");
  dynDynTypes[pos++]  = makeDynInt    (0, DPEL_STRING);

  n = dpTypeChange(dynDynElements, dynDynTypes );

  // check the result of creating dpts
  if(n == -1)
    result = -1;

  dynClear(dynDynElements);
  dynClear(dynDynTypes);

  dynDynElements[1] = makeDynString ("_FwInstallationInformation" , "");
  dynDynElements[2] = makeDynString ("","installationDirectoryPath");
//  dynDynElements[3] = makeDynString ("","postInstallFiles");
//  dynDynElements[4] = makeDynString ("","postDeleteFiles", "");
  dynDynElements[3] = makeDynString ("","lastSourcePath");
  dynDynElements[4] = makeDynString ("","lastInstToolSourcePath");
  dynDynElements[5] = makeDynString ("","addManagersDisabled");
  dynDynElements[6] = makeDynString ("","activateManagersDisabled");
  dynDynElements[7] = makeDynString ("","version");
  dynDynElements[8] = makeDynString ("", "blockUis");
  dynDynElements[9] = makeDynString ("", "deleteFromConfigFile");
  dynDynElements[10] = makeDynString ("", "status");
  dynDynElements[11] = makeDynString ("", "addManagersOnReduPartner");
  dynDynElements[12] = makeDynString ("", "projectInfo");
  dynDynElements[13] = makeDynString ("", "", "wccoaVersion");
  dynDynElements[14] = makeDynString ("", "", "wccoaPatchList");

  dynDynTypes[1] = makeDynInt (DPEL_STRUCT);
  dynDynTypes[2] = makeDynInt (0, DPEL_STRING);
//  dynDynTypes[3] = makeDynInt (0, DPEL_DYN_STRING);
//  dynDynTypes[4] = makeDynInt (0, DPEL_DYN_STRING);
  dynDynTypes[3] = makeDynInt (0, DPEL_STRING);
  dynDynTypes[4] = makeDynInt (0, DPEL_STRING);
  dynDynTypes[5] = makeDynInt (0, DPEL_BOOL);
  dynDynTypes[6] = makeDynInt (0, DPEL_BOOL);
  dynDynTypes[7] = makeDynInt (0, DPEL_STRING);
  dynDynTypes[8] = makeDynInt (0, DPEL_BOOL);
  dynDynTypes[9] = makeDynInt (0, DPEL_BOOL);
  dynDynTypes[10] = makeDynInt (0, DPEL_BOOL);
  dynDynTypes[11] = makeDynInt (0, DPEL_BOOL);
  dynDynTypes[12] = makeDynInt (0, DPEL_STRUCT);
  dynDynTypes[13] = makeDynInt (0, 0, DPEL_STRING);
  dynDynTypes[14] = makeDynInt (0, 0, DPEL_DYN_STRING);

  n = dpTypeChange(dynDynElements, dynDynTypes );

  // check the result of creating dpts
  if(n == -1)
    result = -1;

}

/** This function creates the internal dp-types of the installtation tool pending actions
@return result 0 if OK, -1 otherwise
*/
int fwInstallationInit_createPendingActionsDataPointType()
{
  dyn_dyn_string dynDynElements;
  dyn_dyn_int dynDynTypes;

  int result = 0;

  dynDynElements[1] = makeDynString ("_FwInstallationPendingActions");
  dynDynElements[2] = makeDynString ("","postInstallFiles");
  dynDynElements[3] = makeDynString ("","postDeleteFiles");

  dynDynTypes[1] = makeDynInt (DPEL_STRUCT);
  dynDynTypes[2] = makeDynInt (0, DPEL_DYN_STRING);
  dynDynTypes[3] = makeDynInt (0, DPEL_DYN_STRING);

  return dpTypeChange(dynDynElements, dynDynTypes );
}

int fwInstallationInit_createPendingActionsDp()
{
  if (fwInstallationRedu_isRedundant())
  {
    string dpPair1 = fwInstallation_getInstallationPendingActionsDp(1);
    if(!dpExists(dpPair1))
      dpCreate(dpPair1, "_FwInstallationPendingActions");

    string dpPair2 = fwInstallation_getInstallationPendingActionsDp(2);
    if(!dpExists(dpPair2))
      dpCreate(dpPair2, "_FwInstallationPendingActions");
  }
  else
  {
    string dp = fwInstallation_getInstallationPendingActionsDp();
    if(!dpExists(dp))
      return dpCreate(dp, "_FwInstallationPendingActions");
  }
  return 0;

}

/** This function updates the main dp-type (FwInstallationInformation) of the installation tool from previous versions
  @return 0 if OK, -1 if errors
*/
int fwInstallationInit_updateDPT()
{
  string dp = fwInstallation_getInstallationDp();

  //check pending post-installs and post-delete scripts from old dptype:
  fwInstallationInit_updatePendingInstallationActions(); // only when upgrading installation tool from version < 7.0.0 [Aug 2013] (may be removed in the future)

  string probeDp = "fw_InstallationProbing";
  probeDp = fwInstallationRedu_getLocalDp(probeDp);
  dpCreate(probeDp,"_FwInstallationComponents");

  int count = 0;
  while(!dpExists(probeDp))
  {
    delay(0, 100);
    count++;
    if(count == 300)
      fwInstallation_throw("Failed to create internal fwInstallation probe component dp: " + dp + " for 30 seconds. Cannot continue. Please kill the UI and check para privledges or contact support" , "ERROR");
  }

  // If there was a change in data point structure, update it here
  bool postInitRestartNeededDpeExists = dpExists(fwInstallation_getAgentDp() + "." + fwInstallation_getAfterInitRestartNeededDpElem()); // false when upgrading installation tool from version < 7.2.3 [Nov 2014] (may be removed in the future)
  bool synchronizedComponentsDpeExists = dpExists(fwInstallation_getAgentDp() + FW_INSTALLATION_DB_AGENT_SYNC_COMPONENTS_DPE); // false when upgrading installation tool from version < 8.2.0 [Jun 2018]

  if(!postInitRestartNeededDpeExists || !synchronizedComponentsDpeExists)
  {
    if(fwInstallationInit_createAgentDataPointType("_FwInstallation_agentParametrization", true) != 0){
      fwInstallation_throw("Failed to create/modify _FwInstallation_agentParametrization datapoint type, aborting...", "ERROR");
      return -1;
    }

    if(!postInitRestartNeededDpeExists){
      int dpCreateResult = 0;
      fwInstallationInit_createDataPointTypes(FALSE, dpCreateResult); //Not up-to-date: Update dpts now
      if(dpCreateResult){
        fwInstallation_throw("Failed to create/modify _FwInstallationComponents and/or _FwInstallationInformation datapoint type");
        return -1;
      }

      // Set initial value to new data point (restart after init scripts).
      fwInstallation_clearProjectRestartAfterInit();

      dyn_string protectedManagers;
      _fwInstallationManager_getManagersProtectedFromStopping(protectedManagers);
      if (dynlen(protectedManagers) == 0)
      {
        //initialize the protected managers
        protectedManagers = makeDynString("-m gedi", "-f pvss_scripts.lst", FW_INSTALLATION_DB_AGENT_MANAGER_CMD, "-p fwInstallation/fwInstallation.pnl",
                                          "fwInstallation/fwInstallationFakeScript.ctl", "fwInstallation/fwInstallationAgentDBConsistencyChecker.ctl",
                                          "archiv_client.ctl", "calculateState.ctl",
                                          "libs/PVSSBootstrapper/PVSSBootstrapper_insider.ctl");
        _fwInstallationManager_setManagersProtectedFromStopping(protectedManagers);
      }
    }
    // Set initial value to syncronizationComponents local and central dpe (note: it will be an array with ones)
    dyn_bool syncCompsList;
    if(fwInstallationDBAgent_setEnabledSynchronizationComponents(true, syncCompsList) != 0 ||
       fwInstallationDBAgent_setEnabledSynchronizationComponents(false, syncCompsList) != 0)
    {
      fwInstallation_throw("Failed to set initial values for the list of enabled synchronization components in local and/or central mode", "ERROR");
    }

  }
  // Check if _FwInstallationComponents dpt contains element "sourceFilesHashes", if not, then update dpt
  string newComponentsDpe = probeDp + ".sourceFilesHashes";
  if(!dpExists(newComponentsDpe)) // condition meet when upgrading installation tool from version < 8.1.0 [Dec 2017]
  {
    int dpCreateResult = 0;
    fwInstallationInit_createDataPointTypes(FALSE, dpCreateResult); //Not up-to-date: Update dpts now
    if(dpCreateResult)
      return -1;
    //calculate hashes for currently installed components
    dyn_dyn_string installedComponents;
    if(fwInstallation_getInstalledComponents(installedComponents) != 0)
      fwInstallation_throw("Failed to get installed components properties. Cannot calculate hashes of source files. This needs to be done manually later", "ERROR");

    int installedComponentsCount = dynlen(installedComponents);
    for(int i=1;i<=installedComponentsCount;i++)
      if(fwInstallation_calculateComponentSourceFilesHashes(installedComponents[i][1]) != 0)
        fwInstallation_throw("Error when calculating source files hashes for component: " + installedComponents[i][1] +
                             ". Hash calculation needs for this component to be done manually later", "ERROR");
  }

  // Check if the .qtHelpFiles qp exists
  string qtHelpFilesDpe = probeDp + ".qtHelpFiles";
  if(!dpExists(qtHelpFilesDpe)) // condition meet when upgrading installation
                                // tool from version < 8.4.0 [Dec 2019]
  {
    int dpCreateResult = 0;
    fwInstallationInit_createDataPointTypes(FALSE, dpCreateResult);
    if (dpCreateResult)
      return -1;
  }

  if(dpExists(probeDp))
    dpDelete(probeDp);

  return 0;
}

int fwInstallationInit_setAddManagersOnReduPartner(bool addManagersOnReduPartner)
{
  return dpSetWait(fwInstallation_getInstallationDp() + ".addManagersOnReduPartner", addManagersOnReduPartner);
}

/** This function updates the internal dp-type used by the installation tool for the components
*/
fwInstallationInit_updateComponentDps()
{
  string name;
  dyn_string compDps = fwInstallation_getInstalledComponentDps();

  for(int z = 1; z <= dynlen(compDps); z++)
  {
    dpGet(compDps[z] + ".name", name);

    if(name != fwInstallation_dp2name(compDps[z]))
      dpSet(compDps[z] + ".name", fwInstallation_dp2name(compDps[z]));
  }
}

/** This function checks if the DB-agent of the installation tool is installed in the current project
  @return true if the agent is installed, false otherwise
*/
bool fwInstallationInit_isAgentInstalled()
{
  string dp = fwInstallation_getAgentDp();
  string dpAgent = fwInstallation_getAgentRequestsDp();

  if(dpExists(dp + ".managers.stopDistAfterSync") &&
     dpExists(dp + ".db.connection.schemaOwner") &&
     dpExists(dp + ".db.connection.driver") &&
     dpExists(dpAgent + ".managerReconfiguration.manager"))
  {
    // If this is a redundand system, check if fwInstallation is present there
    if(fwInstallationRedu_isRedundant() &&
       (!dpExists(fwInstallation_getAgentRequestsDp(2))))
    {
      return false;
  }

    return true;
  }

  return false;
}

/** This function installs the DB-Agent of the FW Component Installation Tool
  @return 0 if OK, -1 if errors
*/
int fwInstallationInit_installAgent()
{
  int error = 0;
  dyn_string dataPointTypes = dpTypes();
  string oldVersion;
  string newVersion;
  string dp = fwInstallation_getInstallationDp();
  string dpa = fwInstallation_getAgentDp();

  error = fwInstallationInit_createAgentDataPointType("_FwInstallation_agentParametrization", true);

  if(error){
    fwInstallation_throw("fwInstallationInit_installAgent() -> Could not create DPT for FW installation agent");
    return -1;
  }
  if(fwInstallationInit_createAgentDataPointType("_FwInstallation_agentPendingRequests", true))
  {
    fwInstallation_throw("fwInstallationInit_installAgent() -> Could not create DPT for FW installation agent");
    return -1;
  }

  // check whether the fwInstallation_agentParametrization dp exists
  dpa = fwInstallation_getAgentDp();
  if(!dpExists(dpa))
  {
    error = dpCreate(dpa, "_FwInstallation_agentParametrization");
    if(error){
      fwInstallation_throw("fwInstallationInit_installAgent() -> Could not create DP for FW installation agent: Agent parametrization");
      return -1;
    }
    else
    {
      //initialize the protected managers
      dyn_string protectedManagers = makeDynString("-m gedi", "-f pvss_scripts.lst", FW_INSTALLATION_DB_AGENT_MANAGER_CMD, "-p fwInstallation/fwInstallation.pnl",
                                                   "fwInstallation/fwInstallationFakeScript.ctl", "fwInstallation/fwInstallationAgentDBConsistencyChecker.ctl",
                                                   "archiv_client.ctl", "calculateState.ctl", "libs/PVSSBootstrapper/PVSSBootstrapper_insider.ctl");
      _fwInstallationManager_setManagersProtectedFromStopping(protectedManagers);
      fwInstallationDBAgent_setSyncInterval(300);

      // Set initial value to syncronizationComponents local and central dpe (note: it will be an array with ones)
      dyn_bool syncCompsList;
      if(fwInstallationDBAgent_setEnabledSynchronizationComponents(true, syncCompsList) != 0 ||
         fwInstallationDBAgent_setEnabledSynchronizationComponents(false, syncCompsList) != 0)
      {
        fwInstallation_throw("Failed to set initial values for the list of enabled synchronization components in local and/or central mode", "ERROR");
      }
    }
  }
  else
  {
    //this piece of code is necessary to upgrade from versions previous to 4.0.0
    //as the owner was introduced only in this version of the tool.
    string owner, writer;
    dpGet(dpa + ".db.connection.schemaOwner", owner,
          dpa + ".db.connection.username", writer);

    if(owner == "" || writer == owner)
    {
      //Load init file as schema owner may have been defined there:
      fwInstallationInit_loadInitFile();
    }
  }

  if (fwInstallationRedu_isRedundant())
  {
    string dpr = fwInstallation_getAgentRequestsDp(1);
    if(!dpExists(dpr))
    {
      error = dpCreate(dpr, "_FwInstallation_agentPendingRequests");
      if(error){
        fwInstallation_throw("fwInstallationInit_installAgent() -> Could not create DP for FW installation agent: Agent installation requests pair 1");
        return -1;
      }
    }
    dpr = fwInstallation_getAgentRequestsDp(2);
    if(!dpExists(dpr))
    {
      error = dpCreate(dpr, "_FwInstallation_agentPendingRequests");
      if(error){
        fwInstallation_throw("fwInstallationInit_installAgent() -> Could not create DP for FW installation agent: Agent installation requests pair 2");
        return -1;
      }
    }
  }
  else
  {
    string dpr = fwInstallation_getAgentRequestsDp();
    if(!dpExists(dpr))
    {
      error = dpCreate(dpr, "_FwInstallation_agentPendingRequests");
      if(error){
        fwInstallation_throw("fwInstallationInit_installAgent() -> Could not create DP for FW installation agent: Agent installation requests");
        return -1;
      }
    }
  }

  time tSync;

  dpGet(dpa + ".syncInterval:_online.._stime", tSync);

  if(tSync <= makeTime(2000, 1, 1, 1)){
    fwInstallationDBAgent_setSyncInterval(300);
  }
  dyn_int initialStatus;
  initialStatus[FW_INSTALLATION_DB_STATUS_DEV_AND_APPS] = 0; // initialize list with zeros
  dpSet(dpa + ".db.projectStatus", initialStatus);

  return error;
}

/** This function updates the version of the installation tool in the internal datapoint.
  * Tool version is taken from the fwInstallation.ctl library.
  * @return 0 if datapoint was updated successfully, -1 if update failed.
  */
int fwInstallationInit_updateToolVersionInDp()
{
  string versionLib;
  fwInstallation_getToolVersionLocal(versionLib);

  bool updateNeeded;
  string versionDp;
  if(fwInstallation_getToolVersionFromDp(versionDp) != 0){
    updateNeeded = true;
    fwInstallation_throw("fwInstallationInit_updateToolVersionInDp() -> Could not retrieve installation tool" +
                         " version from datapoint, current version from file will be set in dp", "WARNING");
  }else if(versionDp != versionLib){
    updateNeeded = true;
  }
  if(!updateNeeded){
    return 0; // Version in file and in dp are the same, no need to update, exit.
  }

  if(dpSet(fwInstallation_getInstallationDp() + ".version", versionLib) != 0){
    fwInstallation_throw("fwInstallationInit_updateToolVersionInDp -> Failed to update Installation Tool "+
                         "version in datapoint to version=" + versionLib);
    return -1;
  }

  if(versionDp != ""){ // Other version of installation tool was present in dp, display info about update.
    fwInstallation_throw("INFO: Installation tool updated to version: " + versionLib +
                         ". Previous version: " + versionDp, "INFO");
  }
  return 0;
}

/** This function loads the init file for the installation tool
@return 0 if OK, -1 if error
*/
int fwInstallationInit_loadInitFile()
{
  const string asciiFile = getPath(CONFIG_REL_PATH, gFwInstallationInitFile); // try to find fwInstallationInit.config ASCII file
  if(asciiFile == "" || access(asciiFile, R_OK) != 0){
    // fwInstallation_throw("FW Installation Tool init file cannot be found.", "INFO", 10);
    return 0;
  }

  fwInstallation_throw("FW Installation Tool Init file found. Loading now: " + asciiFile, "INFO", 10);

  dyn_string asciiImportOptions;

  // In split mode we need to explicitly connect to only one host,
  // the one that we are currently being executed on
  if(fwInstallationRedu_isRedundant() && fwInstallationRedu_isSplitMode()){
    // Note: we assume that event and data hosts are the same, this will not support some exotic configurations
    // where active data & even are on separate hosts
    string hostname = fwInstallation_getHostname(myReduHost());
    // If this is a remote UI - we don't support such option (init file would be loaded using local file,
    // while we are connected to a remote system)
    if(strtoupper(hostname) != strtoupper(fwInstallation_getHostname())){
      fwInstallation_throw("fwInstallationInit_loadInitFile(): skipping loading of init file (redundand system in split mode, remote UI).", "WARNING", 10);
      return 0;
    }
    dynAppend(asciiImportOptions, "-event " + hostname + ":" + (string)eventPort());
    dynAppend(asciiImportOptions, "-data "  + hostname + ":" + (string)dataPort());
  }

  dynAppend(asciiImportOptions, "-yes"); // Allow update of DP-types

  string infoFile, logFile;
  if(access(PROJ_PATH + LOG_REL_PATH, W_OK) != 0){ // Can we write to the log directory? Adjust command line accordingly
    fwInstallation_throw("fwInstallationInit_loadInitFile() -> Project log directory not writeable. Omitting stderr output.", "warning");
    dynAppend(asciiImportOptions, "-noVerbose");
  }else{
    dynAppend(asciiImportOptions, "-log +stderr");
    dynAppend(asciiImportOptions, "-log -file");
    infoFile = getPath(LOG_REL_PATH) + fwInstallation_getWCCOAExecutable("ascii") + "_info.log";
    logFile = getPath(LOG_REL_PATH) + fwInstallation_getWCCOAExecutable("ascii") + "_log.log";
  }

  string asciiCommand = fwInstallation_getImportAsciiManagerCommand(asciiFile, asciiImportOptions, infoFile, logFile);
  int errCode = system(asciiCommand);
  if(_UNIX && errCode > 128){
    errCode -= 256;
  }
  if(errCode > 0){
    fwInstallation_throw("fwInstallationInit_loadInitFile() -> Warings while importing FW Installation Tool Init file.", "WARNING");
  }else if(errCode < 0){
    fwInstallation_throw("fwInstallationInit_loadInitFile() -> Errors while importing FW Installation Tool Init file.", "ERROR");
  }

  string instDir;
  dpGet(fwInstallation_getInstallationDp() + ".installationDirectoryPath", instDir);
  if(instDir == ""){
    fwInstallation_throw("fwInstallationInit_loadInitFile() -> Installation directory is not specified, adding to project paths skipped.", "WARNING");
    return 0;
  }
  if(access(instDir, F_OK) != 0){
    mkdir(instDir, 777);
  }
  if(access(instDir, W_OK) != 0){
    fwInstallation_throw("fwInstallationInit_loadInitFile() -> Could not create installation directory", "ERROR", 4);
    return -1;
  }
  if(fwInstallation_addProjPath(instDir, 999) != 0){
    fwInstallation_throw("fwInstallationInit_loadInitFile() -> Could not add installation directory " + instDir + " to project paths", "ERROR");
    return -1;
  }
  return 0;
}

/** This function loads the init file for the installation tool
@return 0 if OK, -1 if error
*/
int fwInstallationInit_runInitScript()
{
  int iReturn = 0;
  string script = "";
  dyn_string paths;

  //find asccii file:
  fwInstallation_getProjPaths(paths);
  for(int i = dynlen(paths); i >= 1; i--)
  {
    paths[i] +="/scripts/";
    if(access(paths[i] + gFwInstallationInitScript, R_OK) == 0)
    {
      script = paths[i] + gFwInstallationInitScript;
      break;
    }
  }

  if(script == "")
  {
//    fwInstallation_throw("FW Installation Tool init script cannot be found.", "INFO", 10);
    return 0;
  }

  fwInstallation_throw("FW Installation Tool Init script found. Executing now: " + script, "INFO", 10);
  fwInstallation_evalScriptFile(script, iReturn);

  return iReturn;
}

int fwInstallationInit_updateWCCOAInfo()
{
  dyn_string patches;
  string version = fwInstallation_getPvssVersion(patches);
  string dp = fwInstallation_getInstallationDp();

  return dpSet(dp + ".projectInfo.wccoaVersion", version,
               dp + ".projectInfo.wccoaPatchList", patches);

}

/** Moves pending actions dpes to separate datapoint.
  * Called when upgrading fwInstallation from version < 7.0.0. Note that version 7.0.0 was released in Aug 2013.
  * @note This function can be marked as obsolete. It is required only when upgrading installation tool from version released before Aug 2013.
  */
int fwInstallationInit_updatePendingInstallationActions()
{
  string dp = fwInstallation_getInstallationDp();
  if(dpExists(dp + ".postInstallFiles") ||
     dpExists(dp + ".postDeleteFiles"))
  {
    dyn_string pendingPostInstalls, pendingPostDeletes;

    dpExists(dp + ".postInstallFiles", pendingPostInstalls,
             dp + ".postDeleteFiles", pendingPostDeletes);

    int result = -1;
    fwInstallationInit_createDataPointTypes(true, result);
    fwInstallationInit_createPendingActionsDataPointType();
    fwInstallationInit_createPendingActionsDp();
    string pendingActionsDp = fwInstallation_getInstallationPendingActionsDp();
    dpSet(pendingActionsDp + ".postInstallFiles", pendingPostInstalls,
          pendingActionsDp + ".postDeleteFiles", pendingPostDeletes);
  }
  else
  {
     string pendingActionsDp = fwInstallation_getInstallationPendingActionsDp();
     if (!dpExists(pendingActionsDp))
       fwInstallationInit_createPendingActionsDp();
  }
  return 0;
}

/** This function creates the internal dp-types of the installtation tool
@param type name of the dp-type to be created
@param create flag to indicated if the dp-type has to be overwritten (obsolete, legacy)
@return result 0 if OK, -1 otherwise
*/

int fwInstallationInit_createAgentDataPointType(string type, bool create)
{
   dyn_dyn_string dynDynElements;
   dyn_dyn_int dynDynTypes;

   int result = 0;

  dynDynElements[1] = makeDynString (type , "");
  dynDynTypes[1] = makeDynInt (DPEL_STRUCT);

  if(type == "_FwInstallation_agentParametrization")
  {
    // Names
     dynDynElements[2] = makeDynString ("","db");
       dynDynElements[3] = makeDynString ("","", "connection");
         dynDynElements[4] = makeDynString ("", "", "", "driver");
        dynDynElements[5] = makeDynString ("", "", "", "server");
        dynDynElements[6] = makeDynString ("", "", "", "username");
        dynDynElements[7] = makeDynString ("" , "", "", "password");
        dynDynElements[8] = makeDynString ("" , "", "", "initialized");
        dynDynElements[9] = makeDynString ("" , "", "", "schemaOwner");
       dynDynElements[10] = makeDynString ("", "", "useDB");
       dynDynElements[11] = makeDynString ("", "", "projectStatus");
//     dynDynElements[12] = makeDynString ("", "", "synchronization");
     dynDynElements[12] = makeDynString ("", "lock");
     dynDynElements[13] = makeDynString ("", "restart");
     dynDynElements[14] = makeDynString ("", "syncInterval");
    dynDynElements[15] = makeDynString ("", "managers");
      dynDynElements[16] = makeDynString ("", "", "stopDist");
      dynDynElements[17] = makeDynString ("", "", "stopUIs");
      dynDynElements[18] = makeDynString ("", "", "stopCtrl");
      dynDynElements[19] = makeDynString ("", "", "stopDistAfterSync");
      dynDynElements[20] = makeDynString ("", "", "protected");
    dynDynElements[21] = makeDynString ("", "redundancy");
      dynDynElements[22] = makeDynString ("", "", "installOnlyInSplit");
      dynDynElements[23] = makeDynString ("", "", "restoreRedundancyAfterInstallation");
      dynDynElements[24] = makeDynString ("", "", "splitModeForced");
      dynDynElements[25] = makeDynString ("", "", "pairToKeepAfterSplit");
    dynDynElements[26] = makeDynString ("", fwInstallation_getFileIssuesSyncDpElem());
    dynDynElements[27] = makeDynString ("", fwInstallation_getAfterInitRestartNeededDpElem());
    dynDynElements[28] = makeDynString ("", fwInstallation_getAfterInitRestartRequesterDpElem());
    dynDynElements[29] = makeDynString ("", "synchronizedComponents");
      dynDynElements[30] = makeDynString ("", "", "enabledInLocal");
      dynDynElements[31] = makeDynString ("", "", "enabledInCentral");


    // Types
    dynDynTypes[2] = makeDynInt (0, DPEL_STRUCT);
       dynDynTypes[3] = makeDynInt (0, 0, DPEL_STRUCT);
         dynDynTypes[4] = makeDynInt (0, 0, 0, DPEL_STRING);
         dynDynTypes[5] = makeDynInt (0, 0, 0, DPEL_STRING);
         dynDynTypes[6] = makeDynInt (0, 0, 0, DPEL_STRING);
         dynDynTypes[7] = makeDynInt (0, 0, 0, DPEL_STRING);
         dynDynTypes[8] = makeDynInt (0, 0, 0, DPEL_INT);
         dynDynTypes[9] = makeDynInt (0, 0, 0, DPEL_STRING);
       dynDynTypes[10] = makeDynInt (0, 0, DPEL_INT);
       dynDynTypes[11] = makeDynInt (0, 0, DPEL_DYN_INT);
//    dynDynTypes[12] = makeDynInt (0, 0, DPEL_DYN_INT);
     dynDynTypes[12] = makeDynInt (0, DPEL_INT);
     dynDynTypes[13] = makeDynInt (0, DPEL_INT);
     dynDynTypes[14] = makeDynInt (0, DPEL_INT);
     dynDynTypes[15] = makeDynInt (0, DPEL_STRUCT);
       dynDynTypes[16] = makeDynInt (0, 0, DPEL_INT);
       dynDynTypes[17] = makeDynInt (0, 0, DPEL_INT);
       dynDynTypes[18] = makeDynInt (0, 0, DPEL_INT);
      dynDynTypes[19] = makeDynInt (0, 0, DPEL_INT);
      dynDynTypes[20] = makeDynInt (0, 0, DPEL_DYN_STRING);
    dynDynTypes[21] = makeDynInt (0, DPEL_STRUCT);
      dynDynTypes[22] = makeDynInt (0, 0, DPEL_INT);
      dynDynTypes[23] = makeDynInt (0, 0, DPEL_INT);
      dynDynTypes[24] = makeDynInt (0, 0, DPEL_BOOL);
      dynDynTypes[25] = makeDynInt (0, 0, DPEL_INT);
    dynDynTypes[26] = makeDynInt (0, DPEL_BOOL);
    dynDynTypes[27] = makeDynInt(0, DPEL_INT);
    dynDynTypes[28] = makeDynInt(0, DPEL_STRING);
    dynDynTypes[29] = makeDynInt(0, DPEL_STRUCT);
    dynDynTypes[30] = makeDynInt(0, 0, DPEL_DYN_BOOL);
    dynDynTypes[31] = makeDynInt(0, 0, DPEL_DYN_BOOL);
  }
  else if(type == "_FwInstallation_agentPendingRequests")
  {
    // Names
     dynDynElements[2] = makeDynString ("", "restart");
     dynDynElements[3] = makeDynString ("", "pvssInstallRequests");
     dynDynElements[4] = makeDynString ("", "pvssDeleteRequests");
    dynDynElements[5] = makeDynString ("", "dbInstallRequests");
    dynDynElements[6] = makeDynString ("", "dbDeleteRequests");
    dynDynElements[7] = makeDynString ("", "trigger");
    dynDynElements[8] = makeDynString ("", "execute");
    dynDynElements[9] = makeDynString ("", "msg");
    dynDynElements[10] = makeDynString ("", "managerReconfiguration");
      dynDynElements[11] = makeDynString ("", "", "manager");
      dynDynElements[12] = makeDynString ("", "", "startMode");
      dynDynElements[13] = makeDynString ("", "", "secKill");
      dynDynElements[14] = makeDynString ("", "", "restartCount");
      dynDynElements[15] = makeDynString ("", "", "resetMin");
      dynDynElements[16] = makeDynString ("", "", "commandLine");

    // Types
    dynDynTypes[2] = makeDynInt (0, DPEL_INT);
     dynDynTypes[3] = makeDynInt (0, DPEL_DYN_STRING);
     dynDynTypes[4] = makeDynInt (0, DPEL_DYN_STRING);
     dynDynTypes[5] = makeDynInt (0, DPEL_DYN_STRING);
     dynDynTypes[6] = makeDynInt (0, DPEL_DYN_STRING);
    dynDynTypes[7] = makeDynInt (0, DPEL_INT);
    dynDynTypes[8] = makeDynInt (0, DPEL_INT);
    dynDynTypes[9] = makeDynInt (0, DPEL_STRING);
    dynDynTypes[10] = makeDynInt (0, DPEL_STRUCT);
       dynDynTypes[11] = makeDynInt (0, 0, DPEL_DYN_STRING);
       dynDynTypes[12] = makeDynInt (0, 0, DPEL_DYN_STRING);
       dynDynTypes[13] = makeDynInt (0, 0, DPEL_DYN_INT);
       dynDynTypes[14] = makeDynInt (0, 0, DPEL_DYN_INT);
       dynDynTypes[15] = makeDynInt (0, 0, DPEL_DYN_INT);
       dynDynTypes[16] = makeDynInt (0, 0, DPEL_DYN_STRING);
  }

   result = dpTypeChange(dynDynElements, dynDynTypes);

  dynClear(dynDynElements);
  dynClear(dynDynTypes);

  return result;
}

private bool fwInstallationInit_consolidation_areLibsVersionsConsistent(){
  return (csFwInstallationInitLibVersion == csFwInstallationToolVersion &&
          csFwInstallationInitLibTag == csFwInstallationToolTag);
}

private string fwInstallationInit_mappingObsoleteCfgLibsToString(const mapping &obsoleteConfigLibs){
  string obsoleteCfgLibsString;
  int obsoleteConfigLibsLen = mappinglen(obsoleteConfigLibs);
  for(int i=1;i<=obsoleteConfigLibsLen;i++){
    string section = mappingGetKey(obsoleteConfigLibs, i);
    string libsStr = strjoin(obsoleteConfigLibs[section], "|");
    obsoleteCfgLibsString += section + "#" + libsStr + "#";
  }
  return strrtrim(obsoleteCfgLibsString, "#");
}

private int fwInstallationInit_consolidation_getObsoleteConfigLibs(string configPath, mapping &obsoleteConfigLibs){
  const dyn_string obsoleteFwInstallationLibraries = makeDynString("fwInstallation.ctl", "fwInstallationDB.ctl");

  dyn_string sections = makeDynString("ui", "ctrl");
  int sectionsLen = dynlen(sections);
  for(int i=1;i<=sectionsLen;i++){
    string section = sections[i];
    dyn_string loadedCtrlLibs;
    if(paCfgReadValueList(configPath, section, "LoadCtrlLibs", loadedCtrlLibs) != 0){
      fwInstallation_throw("Failed to read list of libraries loaded with 'LoadCtrlLibs' entry from config file", "ERROR", 28);
      return -1;
    }
    dyn_string oldLibsLoaded = dynIntersect(obsoleteFwInstallationLibraries, loadedCtrlLibs);
    if(dynlen(oldLibsLoaded) > 0){
      obsoleteConfigLibs[section] = oldLibsLoaded;
    }
  }
  return 0;
}

private int fwInstallationInit_consolidation_removeObsoleteConfigLibs(string configPath, const mapping &obsoleteConfigLibs){
  dyn_string removeFromSections = mappingKeys(obsoleteConfigLibs);
  int removeFromSectionsLen = dynlen(removeFromSections);
  bool err;
  for(int i=1;i<=removeFromSectionsLen;i++){
    string section = removeFromSections[i];
    dyn_string libsToRemove = obsoleteConfigLibs[section];
    int libsToRemoveLen = dynlen(libsToRemove);
    for(int j=1;j<=libsToRemoveLen;j++){
      if(paCfgDeleteValue(configPath, section, "LoadCtrlLibs", libsToRemove[j]) != 0){
        err = true;
        fwInstallation_throw("Failed to remove 'LoadCtrlLibs = " + libsToRemove[j] + " entry from the [" + section + "] section of the project config file.", "ERROR", 28);
      }else{
        fwInstallation_throw("LoadCtrlLibs = " + libsToRemove[j] + " entry removed from the [" + section + "] section of the project config file", "INFO", 28);
      }
    }
  }
  if(err){
    fwInstallation_throw("Please try to locate and manually remove reported entries in the project config file or contact support.", "INFO", 28);
    return -1;
  }
  return 0;
}

private fwInstallationInit_consolidation_getObsoletePathsToRemove(dyn_string &filesToRemove, dyn_string &dirsToRemove, dyn_string &dirsToRemoveRecursively){
  const dyn_string obsoleteScriptFiles =
      makeDynString("fwInstallationAgentDBConsistencyChecker.ctl",
                    "fwInstallationExitSplitMode.ctl",
                    "fwInstallationFakeScript.ctl",
                    "fwInstallationGetDpe.ctl",
                    "fwInstallation_projectRegistration.ctl");
  dynAppendConst(filesToRemove, fwInstallationInit_consolidation_filterPathsExistingInProjDirs(SCRIPTS_REL_PATH, obsoleteScriptFiles));

  const dyn_string obsoleteLibFiles =
      makeDynString("fwInstallation.ctl",
                    "fwInstallationDBAgent.ctl",
                    "fwInstallationDBCache.ctl",
                    "fwInstallationDB.ctl",
                    "fwInstallationDBUpgrade.ctl",
                    "fwInstallationDeprecated.ctl",
                    //"fwInstallationFSM.ctl",
                    //"fwInstallationFSMDB.ctl",
                    "fwInstallationManager.ctl",
                    "fwInstallationPackager.ctl",
                    "fwInstallationProjParam.ctl",
                    "fwInstallationRedu.ctl",
                    "fwInstallationUpgrade.ctl",
                    "fwInstallationXml.ctl");
  dynAppendConst(filesToRemove, fwInstallationInit_consolidation_filterPathsExistingInProjDirs(LIBS_REL_PATH, obsoleteLibFiles));

  // get Installation Tool path from the obsolete files added so far to the list (don't want to use getPath() later on some generic dir names like en_US.iso88591/)
  string fwInstallationPath = PROJ_PATH; // fallback;
  if(dynlen(filesToRemove) > 0){
    string fwInstallationFilePath = filesToRemove[1];

    while(_fwInstallation_fileName(fwInstallationFilePath) != "scripts" &&
          _fwInstallation_fileName(fwInstallationFilePath) != ""){
      fwInstallationFilePath = _fwInstallation_baseDir(fwInstallationFilePath);
    }
    if(_fwInstallation_fileName(fwInstallationFilePath) != ""){
      fwInstallationPath = _fwInstallation_baseDir(fwInstallationFilePath);
    }
    fwInstallation_throw("Detected location of old version of Installation Tool: " + fwInstallationPath, "INFO", 28);
  }

  const dyn_string obsoletePictureFiles =
      makeDynString("5.bmp", "71.xpm", "arrow_single_right_16.png", "fwInstallation_question.bmp", "open_22.png");
  dynAppendConst(filesToRemove, fwInstallationInit_consolidation_filterPathsExistingInDir(fwInstallationPath + PICTURES_REL_PATH, obsoletePictureFiles));

  dyn_string obsoleteDbSchemaFiles = getFileNames(fwInstallationPath + CONFIG_REL_PATH, "fwSysStat_*.sql");
  dynAppendConst(obsoleteDbSchemaFiles, getFileNames(fwInstallationPath + CONFIG_REL_PATH, "fw_sys_stat_sas_target*.sql"));
  dynAppendConst(filesToRemove, fwInstallationInit_consolidation_filterPathsExistingInDir(fwInstallationPath + CONFIG_REL_PATH, obsoleteDbSchemaFiles));

  const string obsoleteImageDir = "fwInstallation/";
  dynAppendConst(dirsToRemoveRecursively, fwInstallationInit_consolidation_filterPathsExistingInProjDirs(IMAGES_REL_PATH, makeDynString(obsoleteImageDir)));

  const string obsoleteIsoDir = "en_US.iso88591/";

  const dyn_string obsoleteMsgCatFiles = makeDynString(obsoleteIsoDir + "fwInstallation.cat");
  dynAppendConst(filesToRemove, fwInstallationInit_consolidation_filterPathsExistingInProjDirs(MSG_REL_PATH, obsoleteMsgCatFiles));
  dynAppendConst(dirsToRemove, fwInstallationInit_consolidation_filterPathsExistingInDir(fwInstallationPath + MSG_REL_PATH, obsoleteIsoDir));

  const string obsoleteWebHelpDir = obsoleteIsoDir + "WebHelp/";
  const string obsoleteFwInstallationHelpDir = obsoleteWebHelpDir + "fwInstallation/";
  dynAppendConst(dirsToRemoveRecursively, fwInstallationInit_consolidation_filterPathsExistingInProjDirs(HELP_REL_PATH, obsoleteFwInstallationHelpDir));
  dynAppendConst(dirsToRemove, fwInstallationInit_consolidation_filterPathsExistingInDir(fwInstallationPath + HELP_REL_PATH, obsoleteWebHelpDir));
  dynAppendConst(dirsToRemove, fwInstallationInit_consolidation_filterPathsExistingInDir(fwInstallationPath + HELP_REL_PATH, obsoleteIsoDir));

  const string obsoleteReleaseNotesFile = makeDynString("fwInstallationReleaseNotes.txt");
  dynAppendConst(filesToRemove, fwInstallationInit_consolidation_filterPathsExistingInProjDirs("", obsoleteReleaseNotesFile));
}

private dyn_string fwInstallationInit_consolidation_filterPathsExistingInProjDirs(string relativeProjPath, const dyn_string &obsoletePathsList){
  dyn_string existingObsoletePathsList;
  int obsoletePathsListLen = dynlen(obsoletePathsList);
  for(int i=1;i<=obsoletePathsListLen;i++){
    string obsoletePath = getPath("", relativeProjPath + obsoletePathsList[i]);
    if(obsoletePath != ""){
      dynAppend(existingObsoletePathsList, obsoletePath);
    }
  }
  return existingObsoletePathsList;
}

private dyn_string fwInstallationInit_consolidation_filterPathsExistingInDir(string absoluteDirPath, const dyn_string &relativePathsList){
  dyn_string absolutePathsList;
  int relativePathsListLen = dynlen(relativePathsList);
  for(int i=1;i<=relativePathsListLen;i++){
    string absolutePath = absoluteDirPath + relativePathsList[i];
    if(access(absolutePath, F_OK) == 0){
      dynAppend(absolutePathsList, absolutePath);
    }
  }
  return absolutePathsList;
}

private fwInstallationInit_consolidation_removeObsoletePaths(const dyn_string &filesToRemove, const dyn_string &dirsToRemove, const dyn_string &dirsToRemoveRecursively){
  fwInstallation_cleanupObsoleteFilesFromPath(gFwInstallationComponentName, "", filesToRemove);
  fwInstallation_cleanupObsoleteFilesFromPath(gFwInstallationComponentName, "", dirsToRemoveRecursively, true);
  fwInstallation_cleanupObsoleteFilesFromPath(gFwInstallationComponentName, "", dirsToRemove);
}
